
/*
 * Copyright (C) Xiaozhe Wang (chaoslawful)
 * Copyright (C) Yichun Zhang (agentzh)
 */


#ifndef _NGX_HTTP_LUA_CAPTUREFILTER_H_INCLUDED_
#define _NGX_HTTP_LUA_CAPTUREFILTER_H_INCLUDED_


#include "ngx_http_lua_common.h"


ngx_int_t ngx_http_lua_capture_filter_init(ngx_conf_t *cf);


#endif /* NGX_HTTP_LUA_CAPTUREFILTER_H */

/* vi:set ft=c ts=4 sw=4 et fdm=marker: */
